# Generated by Django 3.2.11 on 2022-04-14 14:41

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0003_alter_rxrefill_unique_together"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="historicalrx",
            name="medication",
        ),
        migrations.RemoveField(
            model_name="rx",
            name="medication",
        ),
        migrations.AddField(
            model_name="historicalrx",
            name="rx_identifier",
            field=models.CharField(default=uuid.uuid4, max_length=36),
        ),
        migrations.AddField(
            model_name="historicalrx",
            name="rx_name",
            field=models.CharField(default="study prescription", max_length=36),
        ),
        migrations.AddField(
            model_name="historicalrxrefill",
            name="refill_identifier",
            field=models.CharField(default=uuid.uuid4, max_length=36),
        ),
        migrations.AddField(
            model_name="rx",
            name="medications",
            field=models.ManyToManyField(to="edc_pharmacy.Medication"),
        ),
        migrations.AddField(
            model_name="rx",
            name="rx_identifier",
            field=models.CharField(default=uuid.uuid4, max_length=36),
        ),
        migrations.AddField(
            model_name="rx",
            name="rx_name",
            field=models.CharField(default="study prescription", max_length=36),
        ),
        migrations.AddField(
            model_name="rxrefill",
            name="refill_identifier",
            field=models.CharField(default=uuid.uuid4, max_length=36),
        ),
    ]



from django import forms
from edc_randomization.site_randomizers import site_randomizers

from ..models import Rx


def get_last_weight():
    return 100


class RxForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["randomizer_name"] = forms.ChoiceField(
            label="Randomizer name",
            choices=site_randomizers.get_as_choices(),
            required=False,
        )

    subject_identifier = forms.CharField(
        label="Subject Identifier",
        required=False,
        widget=forms.TextInput(attrs={"readonly": "readonly"}),
    )

    class Meta:
        model = Rx
        fields = "__all__"



from django.db import models
from edc_crf.model_mixins import CrfModelMixin
from edc_identifier.managers import SubjectIdentifierManager
from edc_model.models import BaseUuidModel
from edc_screening.model_mixins import ScreeningModelMixin
from edc_sites.models import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_schedule.model_mixins import OffScheduleModelMixin, OnScheduleModelMixin
from edc_visit_tracking.model_mixins import VisitModelMixin

from edc_pharmacy.model_mixins import StudyMedicationCrfModelMixin


class SubjectScreening(ScreeningModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectConsent(SiteModelMixin, BaseUuidModel):
    subject_identifier = models.CharField(max_length=25)

    consent_datetime = models.DateTimeField()

    dob = models.DateTimeField(null=True)

    class Meta:
        pass


class SubjectVisit(SiteModelMixin, VisitModelMixin, BaseUuidModel):
    class Meta(VisitModelMixin.Meta, BaseUuidModel.Meta):
        app_label = "edc_pharmacy"


class StudyMedication(
    StudyMedicationCrfModelMixin,
    CrfModelMixin,
    BaseUuidModel,
):

    subject_visit = models.OneToOneField(SubjectVisit, on_delete=models.PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    def run_metadata_rules_for_crf(self):
        pass

    def metadata_update(self, **kwargs):
        pass

    def update_reference_on_save(self):
        return None

    class Meta(CrfModelMixin.Meta, BaseUuidModel.Meta):
        app_label = "edc_pharmacy"


class OnSchedule(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):

    pass


class OffSchedule(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):

    pass



from django.db import models
from django.db.models import PROTECT
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import BaseUuidModel, HistoricalRecords

from .product import Product


class Manager(models.Manager):

    use_in_migrations = True


class Order(NonUniqueSubjectIdentifierFieldMixin, BaseUuidModel):

    order_datetime = models.DateTimeField(verbose_name="Order date/time")

    product = models.ForeignKey(Product, on_delete=PROTECT)

    qty_ordered = models.DecimalField(null=True, blank=False, decimal_places=2, max_digits=10)

    qty_supplied = models.DecimalField(null=True, blank=False, decimal_places=2, max_digits=10)

    objects = Manager()

    history = HistoricalRecords()

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Medication order"
        verbose_name_plural = "Medication orders"



from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_list_data import site_list_data
from edc_utils import get_utcnow

from edc_pharmacy.models import (
    Box,
    ContainerType,
    Formulation,
    FormulationType,
    GenericContainer,
    Location,
    Medication,
    MedicationLot,
    PillBottle,
    Room,
    Route,
    Shelf,
    Units,
    get_location,
    get_room,
    get_shelf,
)


class TestPrescription(TestCase):
    def setUp(self):
        site_list_data.initialize()
        site_list_data.autodiscover()

        self.medication = Medication.objects.create(
            name="METFORMIN",
        )

        self.formulation = Formulation.objects.create(
            medication=self.medication,
            strength=500,
            units=Units.objects.get(name="mg"),
            route=Route.objects.get(display_name="Oral"),
            formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
        )

    def test_build_location(self):

        location = Location.objects.create(name="LOCATION_ONE")
        room = Room.objects.create(name="ROOM_ONE", location=location)
        shelf = Shelf.objects.create(name="SHELF_ONE", room=room)
        box = Box.objects.create(name="BOX_ONE", shelf=shelf)
        container_type = ContainerType.objects.create(name="Bottle")
        for i in [1, 2, 3]:
            GenericContainer.objects.create(
                container_type=container_type, name=f"BOTTLE_{i}", box=box
            )

    def test_add_unpackaged_medicationitem(self):
        container_type = ContainerType.objects.create(name="bottle")
        medication_lot_one = MedicationLot.objects.create(
            lot_no="LOT1111111",
            formulation=self.formulation,
            expiration_date=get_utcnow() + relativedelta(years=1),
        )
        medication_lot_two = MedicationLot.objects.create(
            lot_no="LOT222222",
            expiration_date=get_utcnow() + relativedelta(years=1),
            formulation=self.formulation,
        )

        location = Location.objects.create(name="LOCATION_ONE")
        room = Room.objects.create(name="ROOM_ONE", location=location)
        shelf_one = Shelf.objects.create(name="SHELF_ONE", room=room)
        Shelf.objects.create(name="SHELF_TWO", room=room)
        Shelf.objects.create(name="SHELF_THREE", room=room)
        Shelf.objects.create(name="SHELF_FOUR", room=room)

        box = Box.objects.create(name="BOX_ONE", shelf=shelf_one)
        for i in [1, 2, 3, 4, 5, 6]:
            PillBottle.objects.create(
                container_type=container_type,
                name=f"BOTTLE_{i}",
                medication_lot=medication_lot_one,
                unit_qty=32,
                box=box,
            )
            PillBottle.objects.create(
                container_type=container_type,
                name=f"BOTTLE_{i + 10}",
                medication_lot=medication_lot_two,
                unit_qty=32,
                box=box,
            )
        for obj in PillBottle.objects.all():
            self.assertEqual(str(obj), "Metformin 500Mg Tablet Oral 32 count")

    def test_where_is_the_bottle(self):
        container_type = ContainerType.objects.create(name="bottle")
        medication_lot_one = MedicationLot.objects.create(
            lot_no="LOT1111111",
            formulation=self.formulation,
            expiration_date=get_utcnow() + relativedelta(years=1),
        )

        location = Location.objects.create(name="LOCATION_ONE")
        room = Room.objects.create(name="ROOM_ONE", location=location)
        shelf_one = Shelf.objects.create(name="SHELF_ONE", room=room)

        box = Box.objects.create(name="BOX_ONE", shelf=shelf_one)
        for i in [1, 2, 3, 4, 5, 6]:
            PillBottle.objects.create(
                container_type=container_type,
                name=f"BOTTLE_{i}",
                medication_lot=medication_lot_one,
                unit_qty=32,
                box=box,
            )
        pill_bottle = PillBottle.objects.get(name="BOTTLE_2")

        self.assertTrue(get_location(item=pill_bottle), location)
        self.assertTrue(get_room(item=pill_bottle), room)
        self.assertTrue(get_shelf(item=pill_bottle), shelf_one)



from typing import Tuple

from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ...admin_site import edc_pharmacy_admin

# from ...forms import PillBottleForm
from ...models import PillBottle
from ..model_admin_mixin import ModelAdminMixin


@admin.register(PillBottle, site=edc_pharmacy_admin)
class PillBottleAdmin(ModelAdminMixin, admin.ModelAdmin):
    show_object_tools = True

    # form = PillBottleForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "container_datetime",
                    "formulation",
                    "medication_lot",
                    "max_unit_qty",
                    "unit_qty",
                    "unit_qty_out",
                    "source_container",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_filter: Tuple[str, ...] = (
        "container_datetime",
        "medication_lot__lot_no",
        "formulation__medication__name",
        "created",
        "modified",
    )

    list_display: Tuple[str, ...] = (
        "container_identifier",
        "formulation",
        "medication_lot",
        "unit_qty",
        "unit_qty_out",
        "created",
        "modified",
    )

    search_fields: Tuple[str, ...] = (
        "medication_lot__lot_no",
        "formulation__medication__name",
    )

    ordering: Tuple[str, ...] = ("-expiration_date",)



from dateutil.relativedelta import relativedelta
from django.test import TestCase, tag
from edc_facility import import_holidays
from edc_list_data import site_list_data
from edc_randomization.constants import ACTIVE, PLACEBO
from edc_randomization.site_randomizers import site_randomizers
from edc_randomization.tests.testcase_mixin import TestCaseMixin, all_sites
from edc_sites import add_or_update_django_sites
from edc_utils import get_utcnow

from edc_pharmacy.exceptions import InsufficientQuantityError
from edc_pharmacy.models import (
    Box,
    ContainerType,
    Formulation,
    FormulationType,
    Location,
    Medication,
    MedicationLot,
    PillBottle,
    Room,
    Route,
    Shelf,
    SubjectPillBottle,
    Units,
    repackage,
    repackage_for_subject,
)


class TestPackaging(TestCaseMixin, TestCase):

    import_randomization_list = True
    site_names = [x.name for x in all_sites]
    sid_count_for_tests = 5

    @classmethod
    def setUpTestData(cls):
        import_holidays(test=True)
        add_or_update_django_sites(sites=all_sites)
        if cls.import_randomization_list:
            randomizer_cls = site_randomizers.get("default")
            randomizer_cls.import_list(
                verbose=False, sid_count_for_tests=cls.sid_count_for_tests
            )
        site_list_data.initialize()
        site_list_data.autodiscover()

    def setUp(self):
        self.medication = Medication.objects.create(
            name="METFORMIN",
        )

        self.formulation = Formulation.objects.create(
            medication=self.medication,
            strength=500,
            units=Units.objects.get(name="mg"),
            route=Route.objects.get(display_name="Oral"),
            formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
        )
        self.medication_lot_one = MedicationLot.objects.create(
            lot_no="LOT1111111",
            formulation=self.formulation,
            expiration_date=get_utcnow() + relativedelta(years=1),
            assignment=ACTIVE,
        )
        self.medication_lot_two = MedicationLot.objects.create(
            lot_no="LOT222222",
            formulation=self.formulation,
            expiration_date=get_utcnow() + relativedelta(years=1),
            assignment=PLACEBO,
        )

    @tag("11")
    def test_repack(self):
        container_type = ContainerType.objects.create(name="bottle")

        location = Location.objects.create(name="LOCATION_ONE")
        room = Room.objects.create(name="ROOM_ONE", location=location)
        shelf_one = Shelf.objects.create(name="SHELF_ONE", room=room)
        shelf_two = Shelf.objects.create(name="SHELF_TWO", room=room)
        box_two = Box.objects.create(name="BOX_TWO", shelf=shelf_two)
        box_one = Box.objects.create(name="BOX_ONE", shelf=shelf_one)
        for i in [1, 2, 3, 4, 5, 6]:
            PillBottle.objects.create(
                container_type=container_type,
                name=f"BOTTLE_{i}",
                medication_lot=self.medication_lot_one,
                unit_qty=500,
                box=box_one,
            )

        source_pill_bottle = PillBottle.objects.get(name="BOTTLE_1")

        for _ in [1, 2, 3]:
            _, source_pill_bottle = repackage(
                PillBottle, 128, box=box_two, source_container=source_pill_bottle
            )

        self.assertRaises(
            InsufficientQuantityError,
            repackage,
            PillBottle,
            128,
            box=box_two,
            source_container=source_pill_bottle,
        )

        self.assertEquals(
            source_pill_bottle.unit_qty - source_pill_bottle.unit_qty_out, 500 - (128 * 3)
        )

        for obj in PillBottle.objects.filter(source_container=source_pill_bottle):
            self.assertEquals(obj.unit_qty, 128)

    @tag("11")
    def test_repack_for_subject(self):
        container_type = ContainerType.objects.create(name="bottle")
        location = Location.objects.create(name="LOCATION_ONE")
        room = Room.objects.create(name="ROOM_ONE", location=location)
        shelf_one = Shelf.objects.create(name="SHELF_ONE", room=room)
        shelf_two = Shelf.objects.create(name="SHELF_TWO", room=room)
        box_two = Box.objects.create(name="BOX_TWO", shelf=shelf_two)
        box_one = Box.objects.create(name="BOX_ONE", shelf=shelf_one)
        for i in [1, 2, 3, 4, 5, 6]:
            PillBottle.objects.create(
                container_type=container_type,
                name=f"BOTTLE_{i}",
                medication_lot=self.medication_lot_one,
                unit_qty=500,
                box=box_one,
            )

        source_pill_bottle = PillBottle.objects.get(name="BOTTLE_1")

        subject_pill_bottle, source_pill_bottle = repackage_for_subject(
            new_container_model_cls=SubjectPillBottle,
            unit_qty=32,
            randomizer_name="default",
            rando_sid="1001",
            subject_identifier=None,
            source_container=source_pill_bottle,
            box=box_two,
        )

        self.assertEqual(source_pill_bottle.unit_qty_out, 32)
        self.assertEqual(subject_pill_bottle.unit_qty, 32)



from django.db import models
from django.db.models import PROTECT
from edc_model import models as edc_models

from .product import Product


class Manager(models.Manager):

    use_in_migrations = True


class StockReceiving(edc_models.BaseUuidModel):

    product = models.ForeignKey(Product, on_delete=PROTECT)

    qty = models.IntegerField()

    stock_identifiers = models.TextField()

    received = models.BooleanField(default=False)

    received_datetime = models.DateTimeField(null=True, blank=True)

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return f"{self.product}: {self.qty} recv'd on {self.received_datetime}"

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Medication stock: Receiving"
        verbose_name_plural = "Medication stock: Receiving"



from __future__ import annotations

from typing import Any

from .refill_creator import RefillCreator, RefillCreatorError


def create_next_refill(instance: Any, related_visit_model_attr: str) -> Any | None:
    """Creates the next refill relative to the current visit,
    if not already created.

    Called from signal.

    Instance should be a subclass of StudyMedicationCrfModelMixin
    """
    rx_refill = None
    refill_end_datetime = None
    if not getattr(instance, "creates_refills_from_crf", None):
        raise RefillCreatorError("Expected an instance of StudyMedicationCrfModelMixin")

    appointment = getattr(instance, related_visit_model_attr).appointment
    if next_appointment := appointment.next:
        if next_next_appointment := next_appointment.next:
            refill_end_datetime = next_next_appointment.appt_datetime

    if refill_end_datetime:
        refill_creator = RefillCreator(
            dosage_guideline=instance.next_dosage_guideline,
            formulation=instance.next_formulation,
            make_active=False,
            refill_start_datetime=instance.refill_end_datetime,
            refill_end_datetime=refill_end_datetime,
            roundup_divisible_by=instance.roundup_divisible_by,
            subject_identifier=getattr(instance, related_visit_model_attr).subject_identifier,
            weight_in_kgs=getattr(instance, "weight_in_kgs", None),
        )
        rx_refill = refill_creator.rx_refill
    return rx_refill



from django.db import models
from django.db.models import PROTECT
from django_crypto_fields.fields import EncryptedCharField
from edc_model.models import BaseUuidModel, HistoricalRecords

from .formulation import Formulation


class Manager(models.Manager):

    use_in_migrations = True


class MedicationLot(BaseUuidModel):

    lot_no = models.CharField(max_length=50, unique=True)

    assignment = EncryptedCharField(null=True)

    expiration_date = models.DateField()

    formulation = models.ForeignKey(Formulation, on_delete=PROTECT)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return f"{self.formulation} Lot {self.lot_no}"

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Medication lot"
        verbose_name_plural = "Medication lots"



from uuid import uuid4

from django.db import models
from edc_model.models import BaseUuidModel
from edc_utils import get_utcnow

from .location import Location


class Room(BaseUuidModel):

    room_identifier = models.CharField(max_length=36, default=uuid4, unique=True)

    room_datetime = models.DateTimeField(default=get_utcnow)

    name = models.CharField(max_length=25, unique=True)

    description = models.TextField(null=True)

    location = models.ForeignKey(Location, on_delete=models.PROTECT)

    def __str__(self):
        return f"Room {self.name}"

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Room"
        verbose_name_plural = "Rooms"



# Generated by Django 3.2.13 on 2022-07-06 20:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0007_auto_20220704_1841"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="historicalrxrefill",
            name="frequency",
        ),
        migrations.RemoveField(
            model_name="historicalrxrefill",
            name="frequency_units",
        ),
        migrations.RemoveField(
            model_name="rxrefill",
            name="frequency",
        ),
        migrations.RemoveField(
            model_name="rxrefill",
            name="frequency_units",
        ),
        migrations.AddField(
            model_name="historicalrxrefill",
            name="round_dose",
            field=models.IntegerField(
                default=0, help_text="Rounds the dose. e.g. 7.3->7.0, 7.5->8.0"
            ),
        ),
        migrations.AddField(
            model_name="historicalrxrefill",
            name="roundup_dose",
            field=models.BooleanField(
                default=False, help_text="Rounds UP the dose. e.g. 7.3->8.0, 7.5->8.0"
            ),
        ),
        migrations.AddField(
            model_name="rxrefill",
            name="round_dose",
            field=models.IntegerField(
                default=0, help_text="Rounds the dose. e.g. 7.3->7.0, 7.5->8.0"
            ),
        ),
        migrations.AddField(
            model_name="rxrefill",
            name="roundup_dose",
            field=models.BooleanField(
                default=False, help_text="Rounds UP the dose. e.g. 7.3->8.0, 7.5->8.0"
            ),
        ),
        migrations.AlterField(
            model_name="historicalrxrefill",
            name="roundup_divisible_by",
            field=models.IntegerField(
                default=0,
                help_text="Rounds up the total. For example, 32 would round 112 pills to 128 pills",
            ),
        ),
        migrations.AlterField(
            model_name="historicalrxrefill",
            name="total",
            field=models.DecimalField(
                blank=True,
                decimal_places=4,
                help_text="Total to be dispensed. Leave blank to auto-calculate",
                max_digits=10,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalrxrefill",
            name="weight_in_kgs",
            field=models.DecimalField(
                blank=True,
                decimal_places=1,
                help_text="Defaults to 1.0",
                max_digits=6,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="rxrefill",
            name="roundup_divisible_by",
            field=models.IntegerField(
                default=0,
                help_text="Rounds up the total. For example, 32 would round 112 pills to 128 pills",
            ),
        ),
        migrations.AlterField(
            model_name="rxrefill",
            name="total",
            field=models.DecimalField(
                blank=True,
                decimal_places=4,
                help_text="Total to be dispensed. Leave blank to auto-calculate",
                max_digits=10,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="rxrefill",
            name="weight_in_kgs",
            field=models.DecimalField(
                blank=True,
                decimal_places=1,
                help_text="Defaults to 1.0",
                max_digits=6,
                null=True,
            ),
        ),
    ]



from django.contrib import admin

from ..admin_site import edc_pharmacy_admin
from ..forms import OrderForm
from ..models import Order
from .model_admin_mixin import ModelAdminMixin


@admin.register(Order, site=edc_pharmacy_admin)
class OrderAdmin(ModelAdminMixin, admin.ModelAdmin):

    show_object_tools = True

    # autocomplete_fields = (
    #     "dosage_guideline",
    #     "formulation",
    # )

    form = OrderForm

    # fieldsets = (
    #     (
    #         "Order",
    #         {
    #             "fields": (
    #                 "stock",
    #                 "qty",
    #             ),
    #         },
    #     ),
    #     (
    #         "Order status",
    #         {
    #             "fields": (
    #                 "packed",
    #                 "packed_datetime",
    #                 "shipped",
    #                 "shipped_datetime",
    #                 "received_at_site",
    #                 "received_at_site_datetime",
    #             ),
    #         },
    #     ),
    #     (
    #         "Refill",
    #         {
    #             "fields": (
    #                 "rx",
    #                 "dosage_guideline",
    #                 "formulation",
    #                 "order_datetime",
    #             )
    #         },
    #     ),
    #     audit_fieldset_tuple,
    # )
    #
    # list_display = (
    #     "subject_identifier",
    #     "order_datetime",
    #     "description",
    #     "packed",
    #     "shipped",
    #     "received_at_site",
    # )
    # list_filter = (
    #     "packed",
    #     "shipped",
    #     "received_at_site",
    #     "order_datetime",
    #     "visit_code",
    #     "visit_code_sequence",
    #     "site",
    # )
    # search_fields = (
    #     "id",
    #     "site__id",
    #     "rx__id",
    #     "rx__rando_sid",
    #     "rx__subject_identifier",
    #     "rx__registered_subject__initials",
    #     "dosage_guideline__medication__name",
    # )
    # ordering = (
    #     "rx__subject_identifier",
    #     "-refill_date",
    # )
    #
    # readonly_fields = (
    #     "rx",
    #     "dosage_guideline",
    #     "formulation",
    #     "refill_date",
    #     "number_of_days",
    #     "packed",
    #     "shipped",
    #     "received_at_site",
    # )
    #
    # @staticmethod
    # def description(obj):
    #     dob = obj.rx.registered_subject.dob.strftime(
    #         convert_php_dateformat(settings.SHORT_DATE_FORMAT)
    #     )
    #   return f"{obj.rx.registered_subject.initials} {dob} {obj.rx.registered_subject.gender}"
    #
    # @staticmethod
    # def gender(obj):
    #     return obj.rx.registered_subject.gender
    #
    # @staticmethod
    # def consented(obj):
    #     return obj.rx.registered_subject.consent_datetime.strftime(
    #         convert_php_dateformat(settings.SHORT_DATE_FORMAT)
    #     )
    #
    # @admin.display(description="Subject")
    # def subject_identifier(self, obj=None):
    #     return obj.rx.subject_identifier



from edc_constants.constants import YES

from ...form_validators import StudyMedicationFormValidator


class StudyMedicationModelFormMixin:

    form_validator_cls = StudyMedicationFormValidator

    def clean(self):
        self.update_refill_end_datetime()
        return super().clean()

    def update_refill_end_datetime(self):
        if (
            not self.cleaned_data.get("refill_end_datetime")
            and self.cleaned_data.get("refill_to_next_visit") == YES
        ):
            if next_appt := self.related_visit.appointment.relative_next:
                self.cleaned_data["refill_end_datetime"] = next_appt.appt_datetime



from __future__ import annotations

from edc_randomization.site_randomizers import site_randomizers

from ...exceptions import (
    InsufficientQuantityError,
    PackagingLotNumberMismatchError,
    PackagingSubjectIdentifierMismatchError,
)


def get_location(item):
    return item.box.shelf.room.location


def get_room(item):
    return item.box.shelf.room


def get_shelf(item):
    return item.box.shelf


def repackage(
    new_container_model_cls=None,
    unit_qty: int = None,
    source_container=None,
    box=None,
    **kwargs,
):
    if source_container.unit_qty - (source_container.unit_qty_out + unit_qty) < 0:
        raise InsufficientQuantityError()
    else:
        new_container = new_container_model_cls(
            container_type=source_container.container_type,
            medication_lot=source_container.medication_lot,
            unit_qty=unit_qty,
            unit_qty_out=0,
            source_container=source_container,
            box=box,
            **kwargs,
        )
        new_container.unit_qty = unit_qty
        new_container.save()
        source_container.unit_qty_out += new_container.unit_qty
        source_container.save()
        source_container.refresh_from_db()
    return new_container, source_container


def repackage_for_subject(
    rando_sid: str = None,
    subject_identifier: str = None,
    randomizer_name=None,
    source_container=None,
    **kwargs,
):
    site_randomizers.get(randomizer_name)
    # confirm sid
    randomization_list = (
        site_randomizers.get(randomizer_name).model_cls().objects.get(sid=rando_sid)
    )
    # confirm assignment matches rando
    if randomization_list.assignment != source_container.medication_lot.assignment:
        raise PackagingLotNumberMismatchError(
            f"Lot number / assignment mismatch. Got sid `{rando_sid}`."
        )
    # confirm subject_identifier matches with sid, if randomized
    if randomization_list.allocated and not subject_identifier:
        raise PackagingSubjectIdentifierMismatchError(
            f"Expected subject identifier. Subject is randomized. Got sid `{rando_sid}`."
        )
    elif not randomization_list.allocated and subject_identifier:
        raise PackagingSubjectIdentifierMismatchError(
            "Did not expect subject identifier. SID has not been allocated. "
            f"Got sid `{rando_sid}`."
        )
    if subject_identifier and randomization_list.subject_identifier != subject_identifier:
        raise PackagingSubjectIdentifierMismatchError(
            f"Subject identifier mismatch. Got sid `{rando_sid}`."
        )
    return repackage(
        rando_sid=rando_sid,
        subject_identifier=subject_identifier,
        source_container=source_container,
        **kwargs,
    )



from dateutil.relativedelta import relativedelta
from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase, tag
from edc_constants.constants import FEMALE
from edc_list_data import site_list_data
from edc_randomization.site_randomizers import site_randomizers
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow

from edc_pharmacy.exceptions import PrescriptionAlreadyExists, PrescriptionError
from edc_pharmacy.models import (
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    Route,
    Rx,
    Units,
)
from edc_pharmacy.prescribe import create_prescription


@tag("rx")
class TestPrescription(TestCase):
    def setUp(self):
        site_list_data.initialize()
        site_list_data.autodiscover()
        self.subject_identifier = "12345"
        self.registered_subject = RegisteredSubject.objects.create(
            subject_identifier=self.subject_identifier,
            gender=FEMALE,
            dob=get_utcnow() - relativedelta(years=25),
            registration_datetime=get_utcnow(),
        )
        self.medication = Medication.objects.create(
            name="Flucytosine",
        )

        self.formulation = Formulation.objects.create(
            strength=500,
            units=Units.objects.get(name="mg"),
            route=Route.objects.get(display_name="Oral"),
            formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
        )

        self.dosage_guideline = DosageGuideline.objects.create(
            medication=self.medication,
            dose_per_kg=100,
            dose_units=Units.objects.get(name="mg"),
            frequency=1,
            frequency_units=FrequencyUnits.objects.get(name="day"),
        )

    def test_create_prescription(self):
        obj = Rx.objects.create(
            subject_identifier=self.subject_identifier,
            report_datetime=get_utcnow(),
        )
        obj.medications.add(self.medication)
        obj.save()

    def test_verify_prescription(self):
        obj = Rx.objects.create(
            subject_identifier=self.subject_identifier,
            report_datetime=get_utcnow(),
        )
        obj.medications.add(self.medication)
        obj.verified = True
        obj.verified = get_utcnow()
        obj.save()
        self.assertTrue(obj.verified)

    def test_create_prescripition_from_func(self):
        create_prescription(
            subject_identifier=self.registered_subject.subject_identifier,
            report_datetime=self.registered_subject.registration_datetime,
            medications=[self.medication],
            site=self.registered_subject.site,
        )
        try:
            Rx.objects.get(subject_identifier=self.registered_subject.subject_identifier)
        except ObjectDoesNotExist:
            self.fail("Rx unexpectedly does not exist")

    def test_create_prescripition_already_exists(self):
        create_prescription(
            subject_identifier=self.registered_subject.subject_identifier,
            report_datetime=self.registered_subject.registration_datetime,
            medications=[self.medication],
            site=self.registered_subject.site,
        )
        Rx.objects.get(subject_identifier=self.registered_subject.subject_identifier)
        with self.assertRaises(PrescriptionAlreadyExists):
            create_prescription(
                subject_identifier=self.registered_subject.subject_identifier,
                report_datetime=self.registered_subject.registration_datetime,
                medications=[self.medication],
                site=self.registered_subject.site,
            )

    def test_create_prescripition_from_func_for_rct(self):
        randomizer = site_randomizers.get("default")
        randomizer.import_list(verbose=False, sid_count_for_tests=3)
        site_randomizers.randomize(
            "default",
            subject_identifier=self.registered_subject.subject_identifier,
            report_datetime=get_utcnow(),
            site=self.registered_subject.site,
            user="jasper",
            gender=self.registered_subject.gender,
        )
        create_prescription(
            subject_identifier=self.registered_subject.subject_identifier,
            report_datetime=self.registered_subject.registration_datetime,
            randomizer_name="default",
            medications=[self.medication],
            site=self.registered_subject.site,
        )
        try:
            Rx.objects.get(subject_identifier=self.registered_subject.subject_identifier)
        except ObjectDoesNotExist:
            self.fail("Rx unexpectedly does not exist")

    def test_create_prescripition_from_func_bad_medication(self):
        try:
            create_prescription(
                subject_identifier=self.registered_subject.subject_identifier,
                report_datetime=self.registered_subject.registration_datetime,
                medications=[self.medication, "blah blah"],
                site=self.registered_subject.site,
            )
        except PrescriptionError:
            pass
        else:
            self.fail("PrescriptionError not raised")



from uuid import uuid4

from django.db import models
from django.db.models import PROTECT
from edc_constants.choices import YES_NO
from edc_constants.constants import NO


class PreOrderRefillModelMixin(models.Model):
    # not used
    pre_order_identifier = models.CharField(max_length=36, default=uuid4, editable=False)

    order_or_update_next = models.CharField(
        verbose_name="Order, or update, refill for next scheduled visit?",
        max_length=15,
        choices=YES_NO,
        default=NO,
    )

    # not used
    next_dosage_guideline = models.ForeignKey(
        "edc_pharmacy.DosageGuideline",
        on_delete=PROTECT,
        related_name="next_dosageguideline",
        null=True,
        blank=True,
    )

    # not used
    next_formulation = models.ForeignKey(
        "edc_pharmacy.Formulation",
        on_delete=PROTECT,
        related_name="next_formulation",
        null=True,
        blank=True,
    )

    class Meta:
        verbose_name = "Pre-order Refill"
        verbose_name_plural = "Pre-order Refills"
        abstract = True



from __future__ import annotations

from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING
from uuid import uuid4
from zoneinfo import ZoneInfo

from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from edc_utils import convert_php_dateformat

from ..exceptions import (
    PrescriptionError,
    PrescriptionExpired,
    PrescriptionNotStarted,
    RefillCreatorError,
)
from ..utils import get_rx_model_cls, get_rxrefill_model_cls
from .activate_refill import activate_refill

if TYPE_CHECKING:
    from ..models import DosageGuideline, Formulation, Rx, RxRefill


class RefillCreator:
    def __init__(
        self,
        refill_identifier: str = None,
        subject_identifier: str = None,
        refill_start_datetime: datetime = None,
        refill_end_datetime: datetime = None,
        formulation: Formulation = None,
        dosage_guideline: DosageGuideline = None,
        make_active: bool | None = None,
        force_active: bool | None = None,
        roundup_divisible_by: int | None = None,
        weight_in_kgs: float | Decimal | None = None,
    ):
        """Creates rx_refill"""
        self._next_rx_refill = None
        self._prev_rx_refill = None
        self.refill_identifier = refill_identifier
        self.refill_end_datetime = refill_end_datetime.astimezone(ZoneInfo("UTC"))
        self.subject_identifier = subject_identifier
        self.refill_start_datetime = refill_start_datetime.astimezone(ZoneInfo("UTC"))
        self.formulation = formulation
        self.dosage_guideline = dosage_guideline
        self.roundup_divisible_by = roundup_divisible_by or 0
        self.weight_in_kgs = weight_in_kgs

        if not self.formulation:
            raise RefillCreatorError(
                f"Formulation cannot be None. Refill identifier is {refill_identifier}. "
                f"See {self.subject_identifier}."
            )
        if not self.dosage_guideline:
            raise RefillCreatorError(
                f"Dosage guideline cannot be None. Refill identifier is {refill_identifier}. "
                f"See {self.subject_identifier}."
            )
        if not self.weight_in_kgs:
            self.weight_in_kgs = self.rx.weight_in_kgs
        if self.dosage_guideline.dose_per_kg and not self.weight_in_kgs:
            raise RefillCreatorError(
                "Dosage guideline requires patient's weight in kgs. "
                f"See {self.subject_identifier}."
            )

        self.make_active = True if make_active is None else make_active
        self.force_active = force_active
        self.rx_refill = self.create_or_update()
        if self.make_active:
            activate_refill(self.rx_refill)

    def create_or_update(self) -> RxRefill:
        """Creates / updates and returns a RxRefill."""
        # find first refill on or after this start date
        opts = dict(
            refill_identifier=self.refill_identifier or uuid4(),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            refill_start_datetime=self.refill_start_datetime,
            refill_end_datetime=self.refill_end_datetime,
            weight_in_kgs=self.weight_in_kgs,
            roundup_divisible_by=self.roundup_divisible_by,
        )
        try:
            rx_refill = get_rxrefill_model_cls().objects.get(
                rx=self.rx, refill_start_datetime=self.refill_start_datetime
            )
        except ObjectDoesNotExist:
            if self.prev_rx_refill:
                # found previous rx_refill, update end datetime, number_of_days
                self.prev_rx_refill.refill_end_datetime = (
                    self.refill_start_datetime - relativedelta(minutes=1)
                )
                self.prev_rx_refill.save()
            if self.next_rx_refill:
                refill_end_datetime = (
                    self.next_rx_refill.refill_start_datetime - relativedelta(minutes=1)
                )
                opts.update(refill_end_datetime=refill_end_datetime)
            # create a new rx_refill
            rx_refill = get_rxrefill_model_cls().objects.create(rx=self.rx, **opts)
        else:
            refill_end_datetime = self.refill_end_datetime
            if self.next_rx_refill:
                refill_end_datetime = (
                    self.next_rx_refill.refill_start_datetime - relativedelta(minutes=1)
                )
            opts.update(refill_end_datetime=refill_end_datetime)
            for k, v in opts.items():
                setattr(rx_refill, k, v)
            rx_refill.save()
            rx_refill.refresh_from_db()
        return rx_refill

    @property
    def rx(self) -> Rx:
        """Returns Rx model instance else raises PrescriptionError"""
        opts = dict(
            subject_identifier=self.subject_identifier,
            medications__in=[self.formulation.medication],
        )
        try:
            rx = get_rx_model_cls().objects.get(**opts)
        except ObjectDoesNotExist:
            raise PrescriptionError(f"Subject does not have a prescription. Got {opts}.")
        else:
            if self.refill_start_datetime.date() < rx.rx_date:
                rx_date = rx.rx_date.strftime(convert_php_dateformat(settings.DATE_FORMAT))
                raise PrescriptionNotStarted(
                    f"Subject's prescription not started. Starts on {rx_date}. "
                    f"Got {self.subject_identifier} attempting "
                    f"refill on {self.refill_start_datetime_as_str}."
                )
            elif (
                rx.rx_expiration_date
                and self.refill_start_datetime.date() > rx.rx_expiration_date
            ):
                rx_expiration_date = rx.rx_expiration_date.strftime(
                    convert_php_dateformat(settings.DATE_FORMAT)
                )
                raise PrescriptionExpired(
                    f"Subject prescription has expired. Expired on {rx_expiration_date}. "
                    f"Got {self.subject_identifier} attempting refill "
                    f"on {self.refill_start_datetime_as_str}."
                )
        return rx

    @property
    def prev_rx_refill(self):
        if not self._prev_rx_refill:
            self._prev_rx_refill = (
                get_rxrefill_model_cls()
                .objects.filter(
                    rx=self.rx, refill_start_datetime__lt=self.refill_start_datetime
                )
                .order_by("refill_start_datetime")
                .first()
            )
        return self._prev_rx_refill

    @property
    def next_rx_refill(self):
        if not self._next_rx_refill:
            self._next_rx_refill = (
                get_rxrefill_model_cls()
                .objects.filter(
                    rx=self.rx, refill_start_datetime__gt=self.refill_start_datetime
                )
                .order_by("refill_start_datetime")
                .first()
            )
        return self._next_rx_refill

    @property
    def refill_start_datetime_as_str(self) -> str:
        return self.refill_start_datetime.strftime(
            convert_php_dateformat(settings.DATETIME_FORMAT)
        )



from django.apps import apps as django_apps
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields.admin import ModelAdminAuditFieldsMixin
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin
from edc_appointment.models import Appointment
from edc_dashboard import url_names
from edc_model_admin.mixins import (
    ModelAdminFormAutoNumberMixin,
    ModelAdminFormInstructionsMixin,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    TemplatesModelAdminMixin,
)


class ModelAdminMixin(
    TemplatesModelAdminMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminFormInstructionsMixin,
    ModelAdminFormAutoNumberMixin,
    ModelAdminRevisionMixin,
    ModelAdminAuditFieldsMixin,
    ModelAdminInstitutionMixin,
):
    subject_dashboard_url_name = "subject_dashboard_url"
    subject_listboard_url_name = "subject_listboard_url"

    def get_subject_dashboard_url_name(self):
        return url_names.get(self.subject_dashboard_url_name)

    def get_subject_dashboard_url_kwargs(self, obj):
        appointment = Appointment.objects.get(
            subject_identifier=obj.subject_identifier,
            visit_code=obj.visit_code,
            visit_code_sequence=obj.visit_code_sequence,
        )
        return dict(
            subject_identifier=obj.subject_identifier,
            appointment=appointment.id,
        )

    def dashboard(self, obj=None, label=None):
        opts = self.get_subject_dashboard_url_kwargs(obj)
        appointment_model_cls = django_apps.get_model("edc_appointment.appointment")
        if not appointment_model_cls.objects.get(id=opts.get("appointment")).visit:
            opts.pop("appointment")
        url = reverse(
            self.get_subject_dashboard_url_name(),
            kwargs=opts,
        )
        context = dict(title="Go to subject's dashboard", url=url, label=label)
        return render_to_string("dashboard_button.html", context=context)



from edc_list_data.model_mixins import ListModelMixin


class Container(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Container"
        verbose_name_plural = "Containers"


class FormulationType(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "FormulationType"
        verbose_name_plural = "FormulationTypes"


class FrequencyUnits(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Frequency units"
        verbose_name_plural = "Frequency units"


class Route(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Route"
        verbose_name_plural = "Routes"


class Units(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Units"
        verbose_name_plural = "Units"



from uuid import uuid4

from django.db import models
from django.db.models import PROTECT
from edc_model import models as edc_models
from edc_sites.models import SiteModelMixin

from .product import Product


class Manager(models.Manager):

    use_in_migrations = True


class Stock(SiteModelMixin, edc_models.BaseUuidModel):

    stock_identifier = models.CharField(max_length=36, default=uuid4, unique=True)

    product = models.ForeignKey(Product, on_delete=PROTECT)

    # TODO: location

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return f"{self.stock_identifier}: {self.product} "

    class Meta(SiteModelMixin.Meta, edc_models.BaseUuidModel.Meta):
        verbose_name = "Medication stock"
        verbose_name_plural = "Medication stock"



from django import forms
from edc_crf.modelform_mixins import CrfModelFormMixin

from edc_pharmacy.form_validators import (
    StudyMedicationFormValidator as BaseStudyMedicationFormValidator,
)

from .models import StudyMedication


class StudyMedicationFormValidator(BaseStudyMedicationFormValidator):
    def validate_demographics(self) -> None:
        pass


class StudyMedicationForm(CrfModelFormMixin, forms.ModelForm):

    form_validator_cls = StudyMedicationFormValidator

    def validate_against_consent(self):
        pass

    class Meta:
        model = StudyMedication
        fields = "__all__"
